from json_cpp import JsonObject, JsonList, JsonString, SearchType, SortOrder, NotFoundBehavior
from .util import *
from .coordinates import *
from .location import *
from .shape import Space
from .world import World_implementation, World
from .cell import Cell_group_builder
from datetime import datetime


class Step(JsonObject):
    def __init__(self,
                 time_stamp: float = 0.0,
                 agent_name: str = "",
                 frame: int = 0,
                 location: Location = None,
                 rotation: float = 0.0,
                 data: str = ""):
        self.time_stamp = time_stamp
        self.agent_name = agent_name
        self.frame = frame
        if location is None:
            location = Location(0, 0)
        self.location = location
        self.rotation = rotation
        self.data = JsonString(data)

class Trajectories(JsonList):
    def __init__(self, iterable=None):
        JsonList.__init__(self, iterable=iterable, list_type=Step)

    def filter_initial_bad_readings(self):
        first_frame = 1
        for i, step in enumerate(self):
            if step.frame != 0:
                break
            first_frame = i
        del self[0:first_frame-1]

    def get_step_index_by_time_stamp(self, time_stamp: float, search_type: int = SearchType.Aprox) -> int:
        return self.find_ordered_index(time_stamp, search_type=search_type, key=lambda s: s.time_stamp)

    def get_step_by_time_stamp(self, time_stamp: float, search_type: int = SearchType.Aprox) -> Step:
        try:
            step_index = self.get_step_index_by_time_stamp(time_stamp, search_type)
            return self[step_index]
        except:
            return None

    def get_step_index_by_frame(self, frame: int, search_type: int = SearchType.Exact) -> int:
        return self.find_ordered_index(frame, search_type=search_type, key=lambda s: s.frame)

    def get_step_by_frame(self, frame: int, search_type: int = SearchType.Exact) -> Step:
        try:
            step_index = self.get_step_index_by_frame(frame, search_type=search_type)
            return self[step_index]
        except:
            return None

    def get_segment(self, start_frame=None, end_frame=None, start_time=None, end_time=None):
        segment = Trajectories()
        s = 0
        e = len(self) - 1
        if start_time is not None:
            s = self.get_step_index_by_time_stamp(start_time, search_type=SearchType.Aprox)
            if self[s].time_stamp < start_time:
                s += 1
        if start_frame is not None:
            s = self.get_step_index_by_frame(start_frame, search_type=SearchType.Aprox)
            if self[s].frame < start_frame:
                s += 1

        if end_time is not None:
            e = self.get_step_index_by_time_stamp(end_time, search_type=SearchType.Aprox)
            if self[s].time_stamp > end_time:
                s -= 1

        if end_frame is not None:
            e = self.get_step_index_by_frame(end_frame, search_type=SearchType.Aprox)
            if self[s].frame > end_frame:
                s -= 1

        for step in self[s:e]:
            segment.append(step)
        return segment

    def get_velocities(self) -> dict:
        velocities = {}
        last_locations = {}
        last_time_stamp = {}
        for s in self:
            if s.agent_name not in velocities:
                velocities[s.agent_name] = Velocities()
                velocities[s.agent_name].append(0.0)
            else:
                time_dif = s.time_stamp - last_time_stamp[s.agent_name]
                velocity = s.location.dist(last_locations[s.agent_name]) / time_dif
                velocities[s.agent_name].append(velocity)
            last_locations[s.agent_name] = s.location
            last_time_stamp[s.agent_name] = s.time_stamp
        return velocities

    def get_locations(self) -> dict:
        locations = {}
        for s in self:
            if s.agent_name not in locations:
                locations[s.agent_name] = Location_list()
            locations[s.agent_name].append(s.location)
        return locations

    def get_stops_steps(self, distance_threshold: float = 0.01, stop_time: float = 0.5):
        stops = []
        future = 0
        for current in range(len(self)):
            if current <= future:
                continue
            future = current
            last_future = future
            while future < len(self) and \
                    self[current].location.dist(self[future].location) < distance_threshold:
                #in case of a big break between frames
                if self[future].time_stamp - self[last_future].time_stamp > stop_time:
                    #it repositions itself in the frame before the break and halts
                    future = last_future
                    break
                last_future = future
                future += 1
            if future == len(self):
                break
            if self[current].time_stamp + stop_time < self[future].time_stamp:
                stops.append((self[current], self[future]))
            else:
                future = 0
        return stops

    def get_stops(self, distance_threshold: float = 0.01, stop_time: float = 0.5):
        return [(b.frame, e.frame) for b, e in self.get_stops_steps(distance_threshold, stop_time)]

    def get_filtered_velocities(self, complementary: float = None, outliers: float = None) -> {}:
        avs = self.get_velocities()
        for agent_name in avs:
            if outliers:
                avs[agent_name] = avs[agent_name].outliers_filter(outliers)
            if complementary:
                avs[agent_name] = avs[agent_name].complementary_filter(complementary)
        return avs

    def get_unique_steps(self):
        unique_steps = Trajectories()
        last_locations = {}
        for s in self:
            if s.agent_name not in last_locations or not last_locations[s.agent_name] == s.location:
                unique_steps.append(s)
            last_locations[s.agent_name] = s.location
        return unique_steps

    def get_agent_names(self) -> list:
        agent_names = []
        for s in self:
            if s.agent_name not in agent_names:
                agent_names.append(s.agent_name)
        return agent_names

    def get_agent_trajectory(self, agent_name: str):
        return self.where("agent_name", agent_name)

    def split_by_agent(self) -> dict:
        return self.split_by("agent_name")


class Episode(JsonObject):
    def __init__(self,
                 start_time: datetime = None,
                 time_stamp: float = 0.0,
                 end_time: datetime = None,
                 trajectories: Trajectories = None,
                 captures: JsonList = None,
                 rewards_sequence: Cell_group_builder = None,
                 rewards_time_stamps: JsonList = None):
        if not start_time:
            start_time = datetime.now()
        self.start_time = start_time
        self.time_stamp = time_stamp
        if not end_time:
            end_time = datetime.now()
        self.end_time = end_time
        if trajectories is None:
            trajectories = Trajectories()
        self.trajectories = trajectories
        if captures is None:
            captures = JsonList(list_type=int)
        self.captures = captures
        if rewards_sequence:
            self.rewards_sequence = rewards_sequence
        else:
            self.rewards_sequence = Cell_group_builder()
        if rewards_time_stamps:
            self.rewards_time_stamps = rewards_time_stamps
        else:
            self.rewards_time_stamps = JsonList(list_type=float)

    def clean(self):
        ts = self.trajectories.get_agent_trajectory("predator").where("frame", 0).get("time_stamp")
        if len(ts) > 0:
            min_time_stamp = min(ts)
            new_trajectories = Trajectories()
            for step in self.trajectories:
                if step.frame != 0 or step.agent_name != "predator" or step.time_stamp == min_time_stamp:
                    new_trajectories.append(step)
            self.trajectories = new_trajectories


class Episode_list(JsonList):
    def __init__(self, iterable=None):
        JsonList.__init__(self, iterable=iterable, list_type=Episode)


class Experiment(JsonObject):
    def __init__(self,
                 name: str = "",
                 world_configuration_name: str = "",
                 world_implementation_name: str = "",
                 occlusions: str = "",
                 subject_name: str = "",
                 duration: int = 0,
                 start_time: datetime = None,
                 episodes: Episode_list = None,
                 episode_count: int = 0,
                 rewards_cells: Cell_group_builder = [],
                 rewards_orientations: JsonList = []):
        self.name = name
        self.world_configuration_name = world_configuration_name
        self.world_implementation_name = world_implementation_name
        self.occlusions = occlusions
        self.subject_name = subject_name
        self.duration = duration
        if start_time is None:
            start_time = datetime.now()
        self.start_time = start_time
        if episodes is None:
            episodes = Episode_list()
        self.episodes = episodes
        self.episode_count = episode_count
        if rewards_cells:
            self.rewards_cells = rewards_cells
        else:
            self.rewards_cells = Cell_group_builder()
        if rewards_orientations:
            self.rewards_orientations = rewards_orientations
        else:
            self.rewards_orientations = JsonList(list_type=int)

    def clean_all_episodes (self):
        for episode in self.episodes:
            episode.clean()

    @staticmethod
    def get_from_file(file_name: str):
        e = Experiment.parse(open(file_name).read())
        check_type(e, Experiment, "")
        return e

    @staticmethod
    def get_from_url(url: str):
        e = Experiment.parse(get_web_json(resource_uri=url))
        check_type(e, Experiment, "")
        return e

    def transform(self, dst_space_name: str):
        check_type(dst_space_name, str, "incorrect type for dst_space")

        dst_space = World_implementation.get_from_name(self.world_configuration_name, dst_space_name).space

        src_space = World_implementation.get_from_name(self.world_configuration_name, self.world_implementation_name).space
        for episode in self.episodes:
            for step in episode.trajectories:
                step.location = Space.transform_to(step.location, src_space, dst_space)
        self.world_implementation_name = dst_space_name

    def set_name(self, prefix: str = None, suffix: str = None):
        self.name = self.start_time.strftime("%Y%m%d_%OH%M_") + self.subject_name + "_" + self.occlusions
        if prefix:
            self.name = prefix + "_" + self.name
        if suffix:
            self.name = self.name + "_" + suffix

    def get_wrong_origin_episodes(self, origin: Coordinates = Coordinates(-20, 0), agent_name: str = "prey") -> JsonList:
        world = World.get_from_parameters_names(self.world_configuration_name, "canonical")
        wrong_origin_episodes = JsonList(list_type=int)
        for i, episode in enumerate(self.episodes):
            agent_trajectory = episode.trajectories.get_agent_trajectory(agent_name)
            if len(agent_trajectory) == 0:
                continue
            episode_origin_location = agent_trajectory[0].location
            episode_origin_index = world.cells.find(episode_origin_location)
            episode_origin_coordinates = world.cells[episode_origin_index].coordinates
            if not episode_origin_coordinates == origin:
                wrong_origin_episodes.append(i)
        return wrong_origin_episodes

    def get_wrong_goal_episodes(self, goal: Coordinates = Coordinates(20, 0), agent_name: str = "prey") -> JsonList:
        world = World.get_from_parameters_names(self.world_configuration_name, "canonical")
        wrong_goal_episodes = JsonList(list_type=int)
        for i, episode in enumerate(self.episodes):
            agent_trajectory = episode.trajectories.get_agent_trajectory(agent_name)
            if len(agent_trajectory) == 0:
                continue
            episode_goal_location = agent_trajectory[-1].location
            episode_goal_index = world.cells.find(episode_goal_location)
            episode_goal_coordinates = world.cells[episode_goal_index].coordinates
            if not episode_goal_coordinates == goal:
                wrong_goal_episodes.append(i)
        return wrong_goal_episodes

    def get_incomplete_episodes(self, threshold_step_count: int = 30, agent_name: str = "prey") -> JsonList:
        incomplete_episodes = JsonList(list_type=int)
        for i, episode in enumerate(self.episodes):
            agent_trajectory = episode.trajectories.get_agent_trajectory(agent_name)
            if len(agent_trajectory) == 0:
                continue
            if len(agent_trajectory) < threshold_step_count:
                incomplete_episodes.append(i)
        return incomplete_episodes

    def get_broken_trajectory_episodes(self, threshold_distance: float = .1, agent_name: str = "prey") -> JsonList:
        broken_trajectory_episodes = JsonList(list_type=int)
        for i, episode in enumerate(self.episodes):
            agent_trajectory = episode.trajectories.get_agent_trajectory(agent_name)
            last_location = None
            for step in agent_trajectory:
                if last_location is not None:
                    if last_location.dist(step.location) > threshold_distance:
                        broken_trajectory_episodes.append(i)
                        break
                last_location = step.location
        return broken_trajectory_episodes

    def remove_episodes(self, episode_list: JsonList) -> None:
        for episode_index in episode_list:
            self.episodes[episode_index].trajectories = Trajectories()
            self.episodes[episode_index].captures = JsonList(list_type=int)


class SyncStep(JsonList):
    def __init__(self):
        JsonList.__init__(self, list_type=Step, allow_empty=True)

    def __getitem__(self, index):
        if isinstance(index, str):
            return self.find_first(lambda s: s and s.agent_name == index, not_found_behavior=NotFoundBehavior.ReturnNone)
        return JsonList.__getitem__(self, index)

class SyncTrajectories(JsonObject):

    def __init__(self, trajectories: Trajectories, agents_names: list = None):
        self.steps = JsonList(list_type=SyncStep)
        split_trajectories = trajectories.split_by_agent()
        if agents_names:
            self.agents_names = JsonList(list_type=str, iterable=agents_names)
        else:
            self.agents_names = JsonList(list_type=str, iterable=split_trajectories)
        self.min_frame = trajectories[0].frame
        self.max_frame = trajectories[-1].frame
        split_trajectories = trajectories.split_by_agent()
        for frame_number in range(self.min_frame, self.max_frame + 1):
            sync_step = SyncStep()
            for agent in split_trajectories:
                try:
                    agent_step = split_trajectories[agent].get_step_by_frame(frame_number)
                    sync_step.append(agent_step)
                except:
                    sync_step.append(None)
            self.steps.append(sync_step)
        JsonObject.__init__(self)


class Data_sequence(JsonList):
    def __init__(self, iterable=None):
        JsonList.__init__(self, iterable=iterable, list_type=float)

    def complementary_filter(self, a: float): #complementary filter
        if a <= 0 or a >= 1:
            raise ArithmeticError("filter parameter should be > 0 and < 1")
        filtered = self.__class__()
        last = self[0]
        for v in self:
            nv = (a * last + (1 - a) * v)
            last = nv
            filtered.append(nv)
        return filtered

    def outliers_filter(self, threshold: float):
        if threshold <= 0:
            raise ArithmeticError("threshold parameter should be > 0")
        threshold = sum(self)/len(self) * threshold
        filtered = self.__class__()
        if len(self) == 0:
            return filtered
        last = self[0]
        for v in self:
            if abs(v-last) <= threshold:
                filtered.append(v)
                last = v
            else:
                filtered.append(last)
        return filtered

    def moving_average_filter(self, window_length: int = 2):
        if window_length <= 0:
            raise ArithmeticError("window length parameter should be > 0")
        filtered = self.__class__()
        if len(self) == 0:
            return filtered
        le = len(self) - window_length
        for i in range(len(self)):
            b = i - window_length if i > window_length else 0
            e = i + window_length if i < le else len(self)
            filtered.append(sum(self[b:e])/(e-b))
        return filtered

    def mean(self):
        return sum(self) / len(self)

    def median(self):
        l = [x for x in self]
        l.sort()
        return l[len(self) // 2]

    def histogram(self, bins=20):
        min_v = min(self)
        max_v = max(self)
        spread = max_v - min_v
        bin_s = spread / bins
        r = JsonList(list_type=int, iterable=[0 for x in range(bins)])
        for v in self:
            i = int((v - min_v) / bin_s)
            if i == bins:
                i = bins - 1
            r[i] += 1
        return r

    def std_err(self):
        from scipy.stats import sem
        return sem(self)

    def std(self):
        from scipy.stats import tstd
        return tstd(self)

    def best_fit(self):
        import numpy as np
        a, b = np.polyfit(range(len(self)), self, 1)
        return Data_sequence([a*x+b for x in range(len(self))])


class Distances(Data_sequence):
    def __init__(self, trajectory: Trajectories = None, trajectory2: Trajectories = None, locations: Location_list = None, locations2: Location_list = None, incremental: bool = False):

        Data_sequence.__init__(self)

        if trajectory:
            locations = trajectory.process(lambda s: s.location)

        if trajectory2:
            locations2 = trajectory2.process(lambda s: s.location)

        if locations is None:
            raise RuntimeError("No input was provided")

        pl = locations[0]
        d = 0
        for i, l in enumerate(locations):
            if locations2:
                self.append(l.dist(locations2[i]))
            else:
                if incremental:
                    d = d + pl.dist(l)
                else:
                    d = pl.dist(l)
                self.append(d)
                pl = l


class Velocities(Data_sequence):
    def __init__(self, trajectory: Trajectories = None, distances: Distances = None, locations: Location_list = None, time_stamps: Data_sequence = None, time_diff: Data_sequence = None):
        Data_sequence.__init__(self)

        if trajectory:
            locations = trajectory.process(lambda s: s.location)

        if locations:
            distances = Distances(locations=locations)

        if trajectory:
            time_stamps = trajectory.process(lambda s: s.time_stamp)

        if time_stamps:
            def td(t):
                r = t - td.lts
                td.lts = t
                return r
            td.lts = 0

            time_diff = time_stamps.process(td)

        if time_diff and distances:
            for i, td in enumerate(time_diff):
                if td:
                    self.append(distances[i]/td)
                else:
                    self.append(0.0)



from json_cpp import JsonObject
from .world import World
from .shape import Polygon_list
from .visibility import to_radians, angle_difference, Location_visibility
from .location import Location
from .util import Timer


class Capture_parameters(JsonObject):

    def __init__(self, distance: float = 0.0, angle: float = 0.0, cool_down: float = 0.0):
        self.distance = distance
        self.angle = angle
        self.cool_down = cool_down


class Capture:

    def __init__(self, parameters: Capture_parameters, world: World):
        self.dist_max = parameters.distance * world.implementation.cell_transformation.size
        self.theta_diff_max = to_radians(parameters.angle / 2)
        occlusions_locations = world.cells.occluded_cells().get("location")
        occlusions_polygons = Polygon_list.get_polygons(occlusions_locations, world.configuration.cell_shape.sides, world.implementation.cell_transformation.size / 2, world.implementation.space.transformation.rotation + world.implementation.cell_transformation.rotation)
        self.visibility = Location_visibility(occlusions_polygons)
        self.cool_down = Timer(parameters.cool_down)

    def is_captured(self, predator_location: Location, predator_theta : float, prey_location: Location) -> bool:
        if self.cool_down:
            return False
        if predator_location.dist(prey_location) > self.dist_max:
            return False
        theta = predator_location.atan(prey_location)
        theta_diff, direction = angle_difference(predator_theta, theta)
        if theta_diff > self.theta_diff_max:
            return False
        if self.visibility.is_visible(predator_location, prey_location):
            self.cool_down.reset()
            return True
        else:
            return False



import numpy
import matplotlib
import matplotlib.pyplot
from .world import *
from .experiment import *
from .agent_markers import *
from .QuickBundles import *


def adjust_color_brightness(color, amount:float=1):
    import matplotlib.colors as mc
    import colorsys
    try:
        c = mc.cnames[color]
    except:
        c = color
    c = colorsys.rgb_to_hls(*mc.to_rgb(c))
    return colorsys.hls_to_rgb(c[0], 1 - amount * (1 - c[1]), c[2])


class Display:

    def __init__(self,
                 world: World,
                 fig_size: tuple = (10, 10),
                 padding: float = .1,
                 outline: float = .5,
                 show_axes: bool = False,
                 cell_color="white",
                 occlusion_color="black",
                 background_color="white",
                 habitat_color="white",
                 cell_edge_color="lightgray",
                 cell_fill: bool = True,
                 show_cell: bool = True,
                 show_occluded_cell: bool = True,
                 show_habitat: bool = True,
                 habitat_fill: bool = True,
                 occluded_cell_edge_color="lightgray",
                 habitat_edge_color="gray",
                 cell_outline_zorder: int = -8,
                 cell_zorder: int = -7,
                 occluded_cell_zorder: int = -7,
                 habitat_zorder: int = -9,
                 cell_alpha: int = 1,
                 cell_outline_alpha: int = 1,
                 habitat_alpha: int = 1,
                 animated: bool = False,
                 ax=None,
                 fig=None):
        if animated:
            matplotlib.pyplot.ion()
        self.cell_alpha = cell_alpha
        self.show_cell = show_cell
        self.show_occluded_cell = show_occluded_cell
        self.show_habitat = show_habitat
        self.cell_outline_alpha = cell_outline_alpha
        self.habitat_alpha = habitat_alpha
        self.cell_fill = cell_fill
        self.habitat_fill = habitat_fill
        self.agents = dict()
        self.agents_markers = dict()
        self.agents_markers["predator"] = Agent_markers.robot()
        self.agents_markers["prey"] = Agent_markers.mouse()
        self.animated = animated
        self.world = world
        if not fig:
            self.fig = matplotlib.pyplot.figure(figsize=fig_size)
        else:
            self.fig = fig
        if not ax:
            self.ax = self.fig.add_subplot(111)
        else:
            self.ax = ax
        self.ax.axes.xaxis.set_visible(show_axes)
        self.ax.axes.yaxis.set_visible(show_axes)
        self.agents_trajectories = Trajectories()
        self.agents_markers = dict()
        self.cells = []
        self.occlusion_color = occlusion_color
        self.habitat_color = habitat_color
        self.habitat_edge_color = habitat_edge_color
        self.cell_color = cell_color
        self.cell_edge_color = cell_edge_color
        self.occluded_cell_edge_color = occluded_cell_edge_color
        self.xcenter = world.implementation.space.center.x
        self.ycenter = world.implementation.space.center.y
        self.outline = outline
        hsize = world.implementation.space.transformation.size / 2
        pad = hsize * padding

        xmin = self.xcenter - hsize - pad
        xmax = self.xcenter + hsize + pad

        ymin = self.ycenter - hsize - pad
        ymax = self.ycenter + hsize + pad

        self.ax.set_xlim(xmin=xmin, xmax=xmax)
        self.ax.set_ylim(ymin=ymin, ymax=ymax)
        self.ax.set_facecolor(background_color)
        self.habitat_theta = math.radians(0 - world.implementation.space.transformation.rotation)
        self.cells_theta = math.radians(0 - world.implementation.cell_transformation.rotation) + self.habitat_theta
        self.cells_size = world.implementation.cell_transformation.size / 2
        self.habitat_size = hsize
        self.cell_polygons = []
        self.cell_outline_polygons = []
        self.habitat_polygon = None
        self.background_extent = [self.xcenter - hsize, self.xcenter + hsize, self.ycenter - hsize, self.ycenter + hsize]
        self.cell_outline_zorder = cell_outline_zorder
        self.cell_zorder = cell_zorder
        self.occluded_cell_zorder = occluded_cell_zorder
        self.habitat_zorder = habitat_zorder
        self.__draw_cells__()
        self.on_cell_click = None
        self.on_background_click = None
        self.on_click = None
        self.on_key_press = None
        self.on_mouse_move = None
        self.track_click = False
        matplotlib.pyplot.tight_layout()

    def __process_click__(self, event):
        if event.xdata is None or event.ydata is None:
            return
        location = Location(event.xdata, event.ydata)
        cell_id = self.world.cells.find(location)
        cell = self.world.cells[cell_id]
        if cell.location.dist(location) < self.world.implementation.cell_transformation.size / 2:
            if self.on_cell_click:
                self.on_cell_click(event.button, cell)
        elif self.on_background_click:
            self.on_background_click(event.button, location)
        if self.on_click:
            self.on_click(event.button)

    def __process_mouse_move__(self, event):
        if event.xdata is None or event.ydata is None:
            return
        location = Location(event.xdata, event.ydata)
        if self.on_mouse_move:
            cell_id = self.world.cells.find(location)
            cell = self.world.cells[cell_id]
            if cell.location.dist(location) < self.world.implementation.cell_transformation.size / 2:
                self.on_mouse_move(location, cell)
            else:
                self.on_mouse_move(location)

    def set_key_pressed_event(self, callback):
        self.on_key_press = callback
        self.fig.canvas.mpl_connect('key_press_event', self.on_key_press)
        matplotlib.pyplot.rcParams['keymap.save'].remove('s')
        matplotlib.pyplot.rcParams['keymap.quit'].remove('q')

    def set_cell_clicked_event(self, callback):
        if not self.track_click:
            self.fig.canvas.mpl_connect('button_press_event', self.__process_click__)
            self.track_click = True
        self.on_cell_click = callback

    def set_background_clicked_event(self, callback):
        if not self.track_click:
            self.fig.canvas.mpl_connect('button_press_event', self.__process_click__)
            self.track_click = True
        self.on_background_click = callback

    def set_clicked_event(self, callback):
        if not self.track_click:
            self.fig.canvas.mpl_connect('button_press_event', self.__process_click__)
            self.track_click = True
        self.on_click = callback

    def set_mouse_move_event(self, callback):
        self.fig.canvas.mpl_connect('motion_notify_event', self.__process_mouse_move__)
        self.track_mouse_move = True
        self.on_mouse_move = callback


    def __draw_cells__(self):
        [p.remove() for p in reversed(self.ax.patches)]
        for cell in self.world.cells:
            color = self.occlusion_color if cell.occluded else self.cell_color
            edge_color = self.occluded_cell_edge_color if cell.occluded else self.cell_edge_color
            fill = True if cell.occluded else self.cell_fill
            visible = self.show_occluded_cell if cell.occluded else self.show_cell
            self.cell_outline_polygons.append(self.ax.add_patch(matplotlib.patches.RegularPolygon(xy=(cell.location.x, cell.location.y),
                                                                               numVertices=self.world.configuration.cell_shape.sides,
                                                                               radius=self.cells_size,
                                                                               facecolor=color,
                                                                               edgecolor=edge_color,
                                                                               orientation=self.cells_theta,
                                                                               zorder=self.cell_outline_zorder,
                                                                               linewidth=1,
                                                                               alpha=self.cell_outline_alpha,
                                                                               fill=fill,
                                                                               visible=visible)))
            self.cell_polygons.append(self.ax.add_patch(matplotlib.patches.RegularPolygon(xy=(cell.location.x, cell.location.y),
                                                                       numVertices=self.world.configuration.cell_shape.sides,
                                                                       radius=self.cells_size * self.outline,
                                                                       facecolor=color,
                                                                       orientation=self.cells_theta,
                                                                       zorder=self.cell_zorder,
                                                                       linewidth=0,
                                                                       alpha=self.cell_alpha,
                                                                       fill=fill,
                                                                       visible=visible)))
        self.habitat_polygon = self.ax.add_patch(matplotlib.patches.RegularPolygon(xy=(self.xcenter, self.ycenter),
                                                                numVertices=self.world.implementation.space.shape.sides,
                                                                radius=self.habitat_size,
                                                                facecolor=self.habitat_color,
                                                                edgecolor=self.habitat_edge_color,
                                                                orientation=self.habitat_theta,
                                                                zorder=self.habitat_zorder,
                                                                alpha=self.habitat_alpha,
                                                                fill=self.habitat_fill,
                                                                visible=self.show_habitat))

    def set_occlusions(self, occlusions: Cell_group_builder):
        self.world.set_occlusions(occlusions)
        self.__draw_cells__()

    def set_agent_marker(self, agent_name: str, marker: matplotlib.path.Path):
        self.agents_markers[agent_name] = marker

    def add_trajectories(self,
                         trajectories: Trajectories,
                         colors: dict = {"prey": "blue", "predator": "red"},
                         alphas: dict = {"prey": 1, "predator": 1},
                         zorder: int = -6,
                         auto_alpha: bool = False,
                         color_map=matplotlib.pyplot.cm.jet,
                         color_range: tuple = (None, None),
                         distance_equalization: bool = False) -> dict:

        def get_segments(locations):
            segments = []
            previous = locations[0]
            for l in locations[1:]:
                segments.append([[previous.x, previous.y], [l.x, l.y]])
                previous = l
            return segments
        lines = {}
        split_trajectories = trajectories.split_by_agent()
        for agent in split_trajectories:
            color = colors[agent]
            alpha = alphas[agent]
            if distance_equalization:
                agent_trajectory = StreamLine(split_trajectories[agent])
            else:
                agent_trajectory = split_trajectories[agent].get("location")

            segments = get_segments(agent_trajectory)

            if auto_alpha:
                step_alphas = [i/len(segments) for i in range(len(segments))]
            else:
                if isinstance(alpha, list):
                    if len(alpha) - 1 == len(segments):
                        alpha = alpha[1:]

                    if len(alpha) == len(segments):
                        step_alphas = alpha
                    else:
                        print("wrong size for alpha list, assigning 1")
                        step_alphas = [1 for i in range(len(segments))]
                else:
                    step_alphas = [alpha for i in range(len(segments))]

            if isinstance(color, list):
                if len(color) - 1 == len(segments):
                    color = color[1:]

                if len(color) == len(segments):
                    if type(color[0]) is float or type(color[0]) is int:
                        min_velocity = min(color) if color_range[0] is None else color_range[0]
                        max_velocity = max(color) if color_range[1] is None else color_range[1]
                        color = [min_velocity if c < min_velocity else max_velocity if c > max_velocity else c for c in color]
                        step_colors = [color_map(c/max_velocity) for c in color]
                    else:
                        step_colors = [matplotlib.colors.to_rgb(c) for c in color]
                else:
                    print("wrong size for color list, expected", len(segments), ", received", len(color), "assigning blue")
                    step_colors = [matplotlib.colors.to_rgb("blue") for i in segments]
            else:
                step_colors = [matplotlib.colors.to_rgb(color) for i in segments]
            color_with_alpha = [(c[0], c[1], c[2], a) for c, a in zip(step_colors, step_alphas)]
            lc = matplotlib.collections.LineCollection(segments, colors=color_with_alpha, lw=3, zorder=zorder)
            lines[agent] = self.ax.add_collection(lc)
        return lines

    def cell(self, cell: Cell = None, cell_id: int = -1, coordinates: Coordinates = None, color=None, outline_color=None, edge_color=None, alpha=None):
        if color is None:
            color = self.cell_color
        if edge_color is None:
            edge_color = self.cell_edge_color
        if cell is None:
            if cell_id == -1:
                if coordinates is None:
                    raise RuntimeError("a cell, cell_id or coordinates must be provided")
                for c in self.world.cells:
                    if c.coordinates == coordinates:
                        cell = c
                        break
                if cell is None:
                    raise RuntimeError("cell coordinates not found")
            else:
                cell = self.world.cells[cell_id]
        if outline_color is None:
            outline_color = color
        if alpha is None:
            alpha = self.cell_alpha
        self.cell_polygons[cell.id].set_facecolor(color)
        self.cell_outline_polygons[cell.id].set_edgecolor(edge_color)
        self.cell_outline_polygons[cell.id].set_facecolor(outline_color)
        self.cell_outline_polygons[cell.id].set_alpha(alpha)

    def heatmap(self, values: list, color_map=matplotlib.pyplot.cm.Reds, value_range: tuple = None) -> None:
        if value_range:
            minv, maxv = value_range
        else:
            minv, maxv = min(values), max(values)

        if (minv == maxv):
            return
        adjusted_values = [(v-minv)/(maxv-minv) for v in values]
        adjusted_cmap = color_map(adjusted_values)
        for cell_id, color in enumerate(adjusted_cmap):
            if not self.world.cells[cell_id].occluded:
                self.cell(cell_id=cell_id, color=color)

    def circle(self, location: Location, radius: float, color, alpha: float = 1.0, zorder: int = -6):
        circle_patch = matplotlib.pyplot.Circle((location.x, location.y), radius, color=color, alpha=alpha, zorder=zorder)
        return self.ax.add_patch(circle_patch)

    def arrow(self, beginning: Location, ending: Location = None, theta: float = 0, dist: float = 0, color="black", head_width: float = .02, alpha: float = 1.0, line_width: float = 0.001, zorder: int = 4, existing_arrow: matplotlib.patches.FancyArrowPatch = None) -> matplotlib.patches.FancyArrowPatch:
        if ending is None:
            ending = beginning.copy().move(theta=theta, dist=dist)
        length = ending - beginning
        if existing_arrow:
            existing_arrow.set_data(x=beginning.x, y=beginning.y, dx=length.x, dy=length.y, head_width=head_width, width=line_width)
            existing_arrow.set_color(color)
            existing_arrow.set_alpha(alpha)
            return existing_arrow
        else:
            new_arrow = self.ax.arrow(beginning.x, beginning.y, length.x, length.y, color=color, head_width=head_width, length_includes_head=True, alpha=alpha, width=line_width, zorder=zorder)
            new_arrow.animated = self.animated
            return new_arrow

    def line(self, beginning: Location, ending: Location = None, theta: float = 0, dist: float = 0, color="black", alpha: float = 1.0, line_width: float = 0.001, zorder: int = -6, existing_line: matplotlib.patches.FancyArrowPatch = None) -> matplotlib.patches.FancyArrowPatch:
        head_width = 0
        if ending is None:
            ending = beginning.copy().move(theta=theta, dist=dist)
        length = ending - beginning
        if existing_line:
            existing_line.set_data(x=beginning.x, y=beginning.y, dx=length.x, dy=length.y, head_width=0, head_length=0, alpha=alpha, width=line_width)
            existing_line.set_color(color)
            existing_line.set_alpha(alpha)
            return existing_line
        else:
            new_arrow = self.ax.arrow(beginning.x, beginning.y, length.x, length.y, color=color, head_width=0, alpha=alpha, head_length=0, length_includes_head=False, width=line_width, zorder=zorder)
            new_arrow.animated = self.animated
            return new_arrow

    def set_background(self, background, alpha: float = 1):
        self.ax.imshow(background, extent=self.background_extent, alpha=alpha)
        self.__draw_cells__()

    def add_icon(self, location: Location, icon_name: str = "", icon_file_path: str = "", rotation: float = 0, size: float = .05, zorder: int = -5, alpha:float = 1):
        from matplotlib import image
        from scipy import ndimage
        if icon_name:
            icon = image.imread(Agent_markers.icon_folder() + "/" + icon_name + ".png")
        else:
            if icon_file_path:
                icon = image.imread(icon_file_path)
            else:
                raise "Either icon_name or icon_file_path must be provided"
        rotated_img = ndimage.rotate(icon, -rotation)
        self.ax.imshow(rotated_img, extent=Agent_markers.icon_box(location, size, rotation), alpha=alpha, zorder=zorder)

    def agent(self, step: Step = None, agent_name: str = None, location: Location = None, rotation: float = None, color=None, size: float = 40.0,  zorder: int = -6, marker: matplotlib.path.Path=None):
        if step:
            agent_name = step.agent_name
            location = step.location
            rotation = step.rotation

        if not marker:
            if agent_name in self.agents_markers:
                marker = self.agents_markers[agent_name]
            else:
                if agent_name == "predator":
                    marker = Agent_markers.robot()
                else:
                    marker = Agent_markers.mouse()

        if agent_name not in self.agents:
            self.agents[agent_name], = self.ax.plot(location.x, location.y, marker=marker, c=color, markersize=size, zorder=zorder)

        t = matplotlib.transforms.Affine2D().rotate_deg_around(0, 0, -rotation)
        self.agents[agent_name].set_marker(marker.transformed(t))
        self.agents[agent_name].set_xdata(location.x)
        self.agents[agent_name].set_ydata(location.y)
        self.agents[agent_name].set_color(color)

    def plot_clusters(self, clusters, colors: list=["blue", "red"], show_centroids: bool = True, use_alpha: bool = True, zorder: int = -6):
        for sl in clusters.unclustered:
            self.ax.plot([s.x for s in sl], [s.y for s in sl], c="gray", linewidth=2, alpha=.5)

        for cn, cluster in enumerate(clusters):
            distances = cluster.get_distances()
            max_distance = max(distances)
            if use_alpha:
                alpha_values = [1 if max_distance == 0 else (max_distance - d) / max_distance / 2 for d in distances]
            else:
                alpha_values = [1 for d in distances]
            color = adjust_color_brightness(colors[cn % len(colors)], .8)
            for i, sl in enumerate(cluster):
                self.ax.plot([s.x for s in sl], [s.y for s in sl], c=color, linewidth=2, alpha=alpha_values[i], zorder=zorder)

        if show_centroids:
            for cn, cluster in enumerate(clusters):
                color = adjust_color_brightness(colors[cn % len(colors)], 1.2)
                cluster_size = .005 + .01 * (len(cluster) / clusters.streamline_count())
                for s in cluster.centroid:
                    self.circle(s, radius=cluster_size + .003, color="lightgray", alpha=1, zorder=500)
                for s in cluster.centroid:
                    self.circle(s, radius=cluster_size, color=color, alpha=1, zorder=500)

    def update(self):
        if self.animated:
            self.fig.canvas.draw()
            self.fig.canvas.flush_events()
            matplotlib.pyplot.pause(.001)

def LoadVideo(video_path: str) -> list:
    import cv2
    video_reader = cv2.VideoCapture(video_path)
    frames = []
    while video_reader.isOpened():
        ret, video_frame = video_reader.read()
        if video_frame is None:
            break
        frames.append(video_frame)
    return frames


def CropVideo(frames: list, crop=(-980, None, None, None, None, None)) -> list:
    cropped_frames = []
    for frame in frames:
        cropped_frames.append(frame[crop[0]:crop[1], crop[2]:crop[3], crop[4]:crop[5]])
    return cropped_frames


def EpisodeRepresentation(episode: Episode,
                          world: World,
                          frames: list = None,
                          time_stamps: list = None,
                          colors: dict = {"prey": "blue", "predator": "red"},
                          cols: int = 0,
                          figsize=(10, 10),
                          video_path: str = "",
                          video_frames: list = None,
                          icons: dict = {"prey": "mouse", "predator": "robot"},
                          distance_equalization: int = 30,
                          icon_size: float = .065, **kwargs):
    split = episode.trajectories.split_by_agent()
    if video_path:
        video_frames = CropVideo(LoadVideo(video_path))
    else:
        video_frames = None

    selections = frames if frames else time_stamps
    if cols == 0:
        cols = len(selections)
    rows = math.ceil(len(selections)/cols)
    fig = matplotlib.pyplot.figure(figsize=figsize)
    axs = fig.subplots(rows, cols)
    fig.tight_layout()
    matplotlib.pyplot.subplots_adjust(wspace=0, hspace=0)
    ps = 0
    displays = []
    if frames is None:
        frames = []
        pf = 0
        for ts in time_stamps:
            for i, step in enumerate(episode.trajectories[pf:]):
                if step.time_stamp > ts:
                    frames.append(step.frame)
                    break
                pf = i
            else:
                frames.append(step.frame)
    for i, frame in enumerate(frames):
        ax = axs[int(i/cols), i % cols]
        if video_frames:
            displays.append(Display(world, cell_fill=False, background_color="black", habitat_fill=False, show_cell=False, show_occluded_cell=False, ax=ax))
            displays[-1].background_extent = [0, 1, .0435, .95]
            displays[-1].set_background(video_frames[frame], 1)
        else:
            displays.append(Display(world, cell_fill=True, background_color="black", habitat_fill=True, show_cell=True, show_occluded_cell=True, ax=ax))

        trajectories = episode.trajectories.get_segment(start_frame=ps, end_frame=frame)
        displays[-1].add_trajectories(trajectories, colors=colors, auto_alpha=True, distance_equalization=distance_equalization)
        if not video_frames:
            for agent in split:
                step = split[agent].get_step_by_frame(frame)
                if "." in icons[agent]:
                    displays[-1].add_icon(step.location, icon_file_path=icons[agent], rotation=step.rotation, size=icon_size)
                else:
                    displays[-1].add_icon(step.location, icon_name=icons[agent], rotation=step.rotation, size=icon_size)
        ps = frame

    print (frames)

    return fig, displays


from matplotlib.path import Path
from .location import Location, Location_list
from .shape import to_radians


class Agent_markers:

    @staticmethod
    def icon_folder() -> str:
        import pkg_resources
        path = 'files'  # always use slash
        filepath = pkg_resources.resource_filename("cellworld", path)
        return filepath

    @staticmethod
    def icon_box(location:Location, base_size: float, rotation: float):
        corners_angles = [to_radians(45+rotation), to_radians(135+rotation), to_radians(225+rotation), to_radians(315+rotation)]
        corners = Location_list([Location(location.x, location.y).move(ca, base_size) for ca in corners_angles])
        x = corners.get_x()
        y = corners.get_y()
        return [min(x), max(x), min(y), max(y)]

    @staticmethod
    def robot(marker_size: float = 1.0) -> Path:
        verts = [
            (-0.25 * marker_size, -1.0 * marker_size),
            (-0.25 * marker_size, 1.0 * marker_size),
            (0.25 * marker_size, 1.0 * marker_size),
            (0.25 * marker_size, - 1.0 * marker_size),
            (-0.25 * marker_size, - 1.0 * marker_size),

            (-0.25 * marker_size, - .1 * marker_size),
            (-0.45 * marker_size, - .1 * marker_size),
            (-0.45 * marker_size, - .5 * marker_size),
            (-0.55 * marker_size, - .5 * marker_size),
            (-0.55 * marker_size, .5 * marker_size),
            (-0.45 * marker_size, .5 * marker_size),
            (-0.45 * marker_size, .1 * marker_size),
            (-0.25 * marker_size, .1 * marker_size),
            (-0.25 * marker_size, -.1 * marker_size),

            (0.25 * marker_size, -.1 * marker_size),
            (0.45 * marker_size, -.1 * marker_size),
            (0.45 * marker_size, -.5 * marker_size),
            (0.55 * marker_size, -.5 * marker_size),
            (0.55 * marker_size, .5 * marker_size),
            (0.45 * marker_size, .5 * marker_size),
            (0.45 * marker_size, .1 * marker_size),
            (0.25 * marker_size, .1 * marker_size),
            (0.25 * marker_size, -.1 * marker_size),

            (0.5 * marker_size, 1.5 * marker_size),
            (0 * marker_size, 1.0 * marker_size),
            (-0.5 * marker_size, 1.5 * marker_size),
        ]

        codes = [
            Path.MOVETO,
            Path.LINETO,
            Path.LINETO,
            Path.LINETO,
            Path.CLOSEPOLY,
            Path.MOVETO,
            Path.LINETO,
            Path.LINETO,
            Path.LINETO,
            Path.LINETO,
            Path.LINETO,
            Path.LINETO,
            Path.LINETO,
            Path.CLOSEPOLY,
            Path.MOVETO,
            Path.LINETO,
            Path.LINETO,
            Path.LINETO,
            Path.LINETO,
            Path.LINETO,
            Path.LINETO,
            Path.LINETO,
            Path.CLOSEPOLY,
            Path.MOVETO,
            Path.LINETO,
            Path.LINETO,
        ]

        return Path(verts, codes)

    @staticmethod
    def arrow(marker_size: float = 1.0) -> Path:
        verts = [(.0 * marker_size, .5 * marker_size),
                 (.5 * marker_size, -.5 * marker_size),
                 (.0 * marker_size, -.25 * marker_size),
                 (-.5 * marker_size, -.5 * marker_size),
                 (.0 * marker_size, .5 * marker_size)]
        codes = [
            Path.MOVETO,  # begin drawing
            Path.LINETO,  # straight line
            Path.LINETO,  # straight line
            Path.LINETO,  # straight line
            Path.CLOSEPOLY
        ]
        return Path(verts, codes)

    @staticmethod
    def mouse(marker_size: float = 1.0) -> Path:
        verts = [(-.8 * marker_size, .0 * marker_size),
                 (-.3 * marker_size, -1.0 * marker_size),
                 (.0 * marker_size, -1.5 * marker_size),
                 (.3 * marker_size, -1.0 * marker_size),
                 (.3 * marker_size, -1.0 * marker_size),
                 (.8 * marker_size, .0 * marker_size),
                 (2.0 * marker_size, -.1 * marker_size),
                 (.8 * marker_size, 2.0 * marker_size),
                 (.6 * marker_size, .0 * marker_size),
                 (.3 * marker_size, .5 * marker_size),
                 (-0.3 * marker_size, .5 * marker_size),
                 (-.6 * marker_size, .0 * marker_size),
                 (-.8 * marker_size, 2 * marker_size),
                 (-2 * marker_size, -.1 * marker_size),
                 (-.8 * marker_size, .0 * marker_size)]
        codes = [
            Path.MOVETO,  # begin drawing
            Path.LINETO,  # straight line
            Path.CURVE3,  # curve
            Path.CURVE3,  # curve
            Path.LINETO,  # straight line
            Path.LINETO,  # straight line

            Path.CURVE4,  # curve
            Path.CURVE4,  # curve
            Path.CURVE4,  # curve

            Path.CURVE4,  # curve
            Path.CURVE4,  # curve
            Path.CURVE4,  # curve

            Path.CURVE4,  # curve
            Path.CURVE4,  # curve
            Path.CURVE4,  # curve
        ]
        return Path(verts, codes)





from .location import Location, Location_list
from .experiment import Trajectories
from json_cpp import JsonObject, JsonList


def get_ratios(a: float, b: float, v: float) -> tuple:
    r = b - a
    ra = (v - a) / r
    rb = (b - v) / r
    return ra, rb

class DistanceMetric:
    MDF = 0
    HAUS = 1
    AMD = 2


class StreamLine(Location_list):
    class StreamLineType:
        Distance = 0
        Time = 1

    def __init__(self, trajectory: Trajectories = None, streamline_len: int = 100, streamline_type=StreamLineType.Distance, fixed_interval=(-1, -1)):
        Location_list.__init__(self, allow_empty=True)
        self.closest_steps = JsonList(list_type=int)
        self.interval = 0

        if trajectory is None:
            return

        if streamline_type == StreamLine.StreamLineType.Distance:
            last_step_location = trajectory[0].location
            total_dist = 0
            distances = []
            for step in trajectory:
                total_dist += last_step_location.dist(step.location)
                distances.append(total_dist)
                last_step_location = step.location
            streamline_step_dist = total_dist / streamline_len
            self.interval = streamline_step_dist

            interval_start = 0
            interval_end = total_dist
            if fixed_interval[0] >= 0:
                interval_start = fixed_interval[0]
                interval_end = fixed_interval[1]

            streamline_step_dist = (interval_end - interval_start) / streamline_len
            step_index = 0
            for i in range(streamline_len + 1):
                streamline_step_total_dist = interval_start + i * streamline_step_dist
                while step_index < len(trajectory) and round(distances[step_index], 4) < round(streamline_step_total_dist,4):
                    step_index += 1
                if step_index == len(trajectory):
                    self.append(None)
                else:
                    if step_index > 0:
                        prev_ratio, next_ratio = get_ratios(distances[step_index - 1], distances[step_index], streamline_step_total_dist)
                        next_streamline_step = trajectory[step_index-1].location * prev_ratio + trajectory[step_index].location * next_ratio
                    else:
                        next_streamline_step = trajectory[step_index].location
                    self.append(next_streamline_step)
                    self.closest_steps.append(step_index)

        else:
            interval_start = trajectory[0].time_stamp
            interval_end = trajectory[-1].time_stamp
            if fixed_interval[0] >= 0:
                interval_start = fixed_interval[0]
                interval_end = fixed_interval[1]

            streamline_step_time = (interval_end - interval_start) / streamline_len
            step_index = 0
            for i in range(streamline_len + 1):
                streamline_step_total_time = interval_start + i * streamline_step_time
                while step_index < len(trajectory) and round(trajectory[step_index].time_stamp, 4) < round(streamline_step_total_time, 4):
                    step_index += 1
                if step_index == len(trajectory):
                    self.append(None)
                else:
                    if step_index > 0:
                        prev_ratio, next_ratio = get_ratios(trajectory[step_index-1].time_stamp, trajectory[step_index].time_stamp, streamline_step_total_time)
                        next_streamline_step = trajectory[step_index-1].location * prev_ratio + trajectory[step_index].location * next_ratio
                    else:
                        next_streamline_step = trajectory[step_index].location
                    self.append(next_streamline_step)
                    self.closest_steps.append(step_index)

    def distance(self, streamline, distance_metric: DistanceMetric = DistanceMetric.MDF) -> float:
        if distance_metric == DistanceMetric.MDF:
            if len(self) != len(streamline):
                raise RuntimeError("StreamLines must have the same length")
            f_total_distance = 0
            for i, s in enumerate(self):
                f_total_distance += streamline[i].dist(s)

            b_total_distance = 0
            for i, s in enumerate(self[::-1]):
                b_total_distance += streamline[i].dist(s)
                if b_total_distance > f_total_distance:
                    return f_total_distance / len(self)
            return b_total_distance / len(self)
        elif distance_metric == DistanceMetric.HAUS:
            d = float()
            for p in self:
                p_dist = streamline.process(lambda l: l.dist(p))
                d = max(d, min(p_dist))
            return d
        elif distance_metric == DistanceMetric.AMD:
            d = float()
            for p in self:
                p_dist = streamline.process(lambda l: l.dist(p))
                d += min(p_dist)
            return d / len(self)


    @classmethod
    def combine(cls, streamlines):
        combined = StreamLine()
        for s in range(len(streamlines[0])):
            combined_step = Location()
            for sl in streamlines:
                combined_step += sl[s]
            combined_step.x /= len(streamlines)
            combined_step.y /= len(streamlines)
            combined.append(combined_step)
        return combined

    def clone(self):
        r = StreamLine()
        for s in self:
            r.append(Location(s.x, s.y))
        return r


class StreamLineCluster(JsonObject):

    def __iter__(self):
        return self.streamlines.__iter__()

    def __len__(self):
        return len(self.streamlines)

    def __init__(self, streamline: StreamLine = None):
        self.streamlines = JsonList(list_type=StreamLine)
        if streamline:
            self.streamlines.append(streamline)
            self.centroid = streamline
        else:
            self.centroid = StreamLine()
        JsonObject.__init__(self)

    def add_streamline(self, streamline: StreamLine, update_centroid: bool = True) -> None:
        if self.centroid and len(self.centroid) != len(streamline):
            raise RuntimeError("StreamLines must have the same length")
        if update_centroid:
            self.centroid = self.__get_new_centroid__(streamline)
        self.streamlines.append(streamline)

    def __get_new_centroid__(self, streamline: StreamLine) -> StreamLine:
        if not self.centroid:
            return streamline.clone()
        new_centroid = StreamLine()
        streamline_count = len(self.streamlines)
        for i, s in enumerate(self.centroid):
            combined_step = (streamline[i] + s * streamline_count) * (1 / (streamline_count + 1))
            new_centroid.append(combined_step)
        return new_centroid

    def distance(self, streamline: StreamLine,
                 update_centroid=True,
                 distance_metric: DistanceMetric = DistanceMetric.MDF) -> float:
        if update_centroid:
            new_centroid = self.__get_new_centroid__(streamline)
            return new_centroid.distance(streamline, distance_metric)
        else:
            return self.centroid.distance(streamline, distance_metric)

    def get_distances(self,
                      distance_metric: DistanceMetric = DistanceMetric.MDF) -> JsonList:
        distances = JsonList(list_type=float)
        for sl in self.streamlines:
            distances.append(sl.distance(self.centroid, distance_metric))
        return distances


class StreamLineClusters(JsonObject):

    def __iter__(self):
        return self.clusters.__iter__()

    def __len__(self):
        return len(self.clusters)

    def __init__(self,
                 min_clusters: int = 1,
                 max_distance: float = .1,
                 streamline_len: int = 100,
                 distance_metric: DistanceMetric = DistanceMetric.MDF):
        self.distance_metric = distance_metric
        self.min_clusters = min_clusters
        self.max_distance = max_distance
        self.streamline_len = streamline_len
        self.clusters = JsonList(list_type=StreamLineCluster)
        self.unclustered = JsonList(list_type=StreamLine)

    def evaluate_trajectory(self, trajectory: Trajectories) -> StreamLineCluster:
        streamline = StreamLine(trajectory=trajectory, streamline_len=self.streamline_len)
        return self.evaluate_streamline(streamline)

    def get_closest_cluster(self, streamline: StreamLine, update_centroid=True) ->tuple:
        if self.streamline_len + 1 != len(streamline):
            raise RuntimeError("StreamLines must have the same length")
        closest_cluster_distance = 0
        closest_cluster = None
        for c in self.clusters:
            distance = c.distance(streamline,
                                  update_centroid,
                                  distance_metric=self.distance_metric)
            if closest_cluster is None or distance < closest_cluster_distance:
                closest_cluster = c
                closest_cluster_distance = distance
        return closest_cluster, closest_cluster_distance

    def evaluate_streamline(self, streamline: StreamLine) -> StreamLineCluster:
        closest_cluster, closest_cluster_distance = self.get_closest_cluster(streamline, True)
        if closest_cluster_distance < self.max_distance:
            return closest_cluster
        else:
            return None

    def add_streamline(self, streamline: StreamLine) -> None:
        candidate_cluster = self.evaluate_streamline(streamline)
        if candidate_cluster is None:
            self.clusters.append(StreamLineCluster(streamline=streamline))
        else:
            candidate_cluster.add_streamline(streamline)

    def filter_clusters(self, threshold: int) -> None:
        self.unclustered = JsonList(list_type=StreamLine)
        while True:
            for i, c in enumerate(self.clusters):
                if len(c) < threshold:
                    break
            else:
                return
            self.unclustered += c
            self.clusters.remove(c)

    def streamline_count(self) -> int:
        counter = 0
        if self.unclustered:
            counter += len(self.unclustered)
        for c in self.clusters:
            counter += len(c.streamlines)
        return counter

    def add_trajectory(self, trajectory: Trajectories) -> None:
        streamline = StreamLine(trajectory=trajectory, streamline_len=self.streamline_len)
        self.add_streamline(streamline)

    def get_average_distance(self):
        distances = self.get_distances()
        if len(distances) == 0:
            return 0
        return sum(distances)/len(distances)

    def get_distances(self) -> JsonList:
        distances = self.get_intra_cluster_distances()
        for usl in self.unclustered:
            c, d = self.get_closest_cluster(usl, False)
            distances.append(d)
        return distances

    def get_intra_cluster_distances(self) -> JsonList:
        distances = JsonList(list_type=float)
        for c in self.clusters:
            distances += c.get_distances()
        return distances

    def get_average_intra_cluster_distance(self):
        distances = self.get_intra_cluster_distances()
        if len(distances) == 0:
            return 0
        return sum(distances)/len(distances)

    def get_cluster_population_standard_deviation(self):
        count = self.streamline_count()
        values = [len(c)/count for c in self.clusters]
        mean = sum(values)/len(values)
        dev = [abs(v-mean) ** 2 for v in values]
        std_dev = (sum(dev)/len(dev)) ** .5
        return std_dev

    def assign_unclustered(self):
        for u in self.unclustered:
            cc, d = self.get_closest_cluster(u)
            cc.add_streamline(u, update_centroid=False)
        self.unclustered.clear()
        


from json_cpp import JsonObject, JsonList
from .coordinates import Coordinates, Coordinates_list
from .util import get_resource
from .world import World
from .cell import Cell, Cell_map, Cell_group


class Paths_builder(JsonObject):
    def __init__(self):
        self.moves = Coordinates_list()
        self.steps = JsonList(list_type=int)
        JsonObject.__init__(self)

    @staticmethod
    def get_from_name(world_configuration_name: str, occlusions_name: str, path_type_name: str = "astar"):
        if not type(world_configuration_name) is str:
            raise "incorrect type for world_configuration_name"
        if not type(occlusions_name) is str:
            raise "incorrect type for occlusions_name"
        if not type(path_type_name) is str:
            raise "incorrect type for path_type_name"
        return Paths_builder.parse(json_dictionary=get_resource("paths", world_configuration_name, occlusions_name, path_type_name))


class Paths:
    def __init__(self, builder: Paths_builder, world: World):
        self.moves = builder.moves
        self.steps = builder.steps
        self.world = world
        self.map = Cell_map(world.configuration.cell_coordinates)

    def _get_index(self, s_ind, d_ind):
        computed_index = s_ind * len(self.world.cells) + d_ind
        return computed_index

    def get_move(self, src_cell: Cell, dst_cell: Cell):
        move_index = self._get_index(src_cell.id, dst_cell.id)
        move = self.moves[move_index]
        return move

    def get_path(self, src_cell: Cell, dst_cell: Cell):
        rst = Cell_group()
        curr_cell = src_cell
        move = self.get_move(curr_cell, dst_cell)
        curr_coordinates = src_cell.coordinates
        rst.append(curr_cell)
        while move != Coordinates(0, 0) and curr_cell != dst_cell:
            curr_coordinates = curr_coordinates + move
            new_cell_id = self.map[curr_coordinates]
            if new_cell_id == -1:
                break
            curr_cell = self.world.cells[new_cell_id]
            move_index = self._get_index(curr_cell.id, dst_cell.id)
            move = self.moves[move_index]
            rst.append(curr_cell)
        return rst



import math

from scipy.stats import entropy


def probability_entropy(prob, base: float = math.e):
    from math import log
    ent = 0
    for p in prob:
        if p > 0:
            ent += p * log(p, base)
    return -ent


def fair_probability(symbol_count: int):
    return [1 / symbol_count for i in range(symbol_count)]


def fair_entropy(symbol_count: int, base: float = math.e):
    return probability_entropy(fair_probability(symbol_count), base=base)


def histogram(data, labels=None, closest: bool = False):
    hist = {}
    if closest and (labels is None or len(labels) == 0):
        raise RuntimeError("if closest is used then labels must be specified")

    if labels:
        for l in labels:
            hist[l] = 0
    for v in data:
        if closest:
            min_distance = abs(v-labels[0])
            i = labels[0]
            for l in labels[1:]:
                distance = abs(v-labels[0])
                if min_distance > distance:
                    i = l
                    min_distance = distance
        i = v
        if i in hist:
            hist[i] += 1
        else:
            if labels is None:
                hist[i] = 1
            else:
                raise RuntimeError("data with value " + str(i) + " not present in labels")
    return list(hist.keys()), list(hist.values())


def entropy(data, labels=None, base: float = math.e):
    labels, values = histogram(data, labels)
    data_count = sum(values)
    data_probability = [v/data_count for v in values]
    data_entropy = probability_entropy(data_probability, base)
    return data_entropy


def normalized_entropy(data, labels=None, base: float = math.e):
    labels, values = histogram(data, labels)
    data_count = sum(values)
    data_probability = [v/data_count for v in values]
    data_entropy = probability_entropy(data_probability, base=base)
    if data_entropy == 0:
        return 0
    label_count = len(labels)
    if label_count == 1:
        return 0
    fair_entropy = probability_entropy(fair_probability(label_count), base=base)
    return data_entropy / fair_entropy




import math

import networkx as nx
from .info import normalized_entropy, entropy
from .cell import Cell_map, Cell_group, Cell, Cell_group_builder
from .world import World
from .visibility import Location_visibility
from json_cpp import JsonList
from .util import get_resource


class Graph_builder(JsonList):
    def __init__(self):
        JsonList.__init__(self, list_type=Cell_group_builder)

    @staticmethod
    def get_from_name(world_name: str, name: str, *argv):
        if not type(world_name) is str:
            raise "incorrect type for world_name"
        if not type(name) is str:
            raise "incorrect type for name"
        return Graph_builder.parse(json_list=get_resource("graph", world_name, name, *argv))


class Graph:

    @staticmethod
    def create_visibility_graph(world: World, include_self_reference=False):
        graph = Graph(cells=world.cells)
        location_visibility = Location_visibility.from_world(world)
        for src in world.cells:
            if src.occluded:
                continue
            for dst in world.cells:
                if dst == src and not include_self_reference:
                    continue
                if graph.is_connected(src, dst):
                    continue
                if location_visibility.is_visible(src.location, dst.location):
                    graph.connect(src, dst)
        return graph

    @staticmethod
    def create_connection_graph(world: World):
        graph = Graph(cells=world.cells)
        map = Cell_map(world.configuration.cell_coordinates)
        for cell in world.cells:
            src = cell
            src_coord = cell.coordinates
            if not cell.occluded:
                for conn in world.configuration.connection_pattern:
                    dst_coord = src_coord + conn
                    dst_index = map[dst_coord]
                    if dst_index >= 0:
                        dst = world.cells[dst_index]
                        if not dst.occluded:
                            graph.connect(src, dst)
        return graph

    def __init__(self, builder: Graph_builder = None, cells: Cell_group = None):
        if builder:
            self._connections = builder
        else:
            self._connections = Graph_builder()
        self.cells = cells

    def connect(self, src: Cell, dst: Cell, bi=True):
        while src.id >= len(self._connections):
            self._connections.append(Cell_group_builder())
        while dst.id >= len(self._connections):
            self._connections.append(Cell_group_builder())

        if not dst.id in self._connections[src.id]:
            self._connections[src.id].append(dst.id)
        if bi and not src.id in self._connections[dst.id]:
            self._connections[dst.id].append(src.id)

    def is_connected(self, src: Cell, dst: Cell) -> bool:
        if src.id == dst.id:
            return True
        if src.id >= len(self._connections):
            return False
        if dst.id >= len(self._connections[src.id]):
            return False
        pending = [src.id]
        visited = []
        while pending:
            next_id = pending[0]
            if dst.id in self._connections[next_id]:
                return True
            visited.append(next_id)
            pending = pending[1:]
            pending += [c for c in self._connections[next_id] if not c in visited and not c in pending]
        return False

    def to_nxgraph (self, include_occluded=False):
        nxgraph = nx.Graph()
        if self.cells:
            for cell in self.cells:
                if include_occluded or not cell.occluded:
                    nxgraph.add_node(cell.id, pos=(cell.location.x, cell.location.y), color="black" if cell.occluded else "white")
        else:
            nxgraph.add_nodes_from(list(range(len(self._connections))))
        for src, conns in enumerate(self._connections):
            for dst in conns:
                nxgraph.add_edge(src, dst)
        return nxgraph

    def __getitem__(self, cell: Cell) -> Cell_group_builder:
        return self._connections[cell.id]

    def get_builder(self):
        return self._connections

    def vertex_degrees(self, cell: Cell) -> int:
        return len(self[cell])

    def complexity(self, base: float = math.e, normalized: bool = False):
        counters = [len(c) for c in self._connections if c]
        if normalized:
            return normalized_entropy(counters, labels=list(range(1, len(counters))), base=base)
        else:
            return entropy(counters, labels=list(range(1, len(counters))), base=base)

    def get_centrality(self, depth: int = 1) -> JsonList:

        centrality = JsonList(list_type=float, iterable=[1.0 for _ in self.cells])
        for i in range(depth):
            iteration_centrality = JsonList(list_type=float, iterable=[0.0 for _ in self.cells])
            for cell_id in range(len(self.cells)):
                for connection in self._connections[cell_id]:
                    iteration_centrality[cell_id] += centrality[connection]
            total = sum(iteration_centrality)
            for cell_id in range(len(self.cells)):
                centrality[cell_id] = iteration_centrality[cell_id] / total
        return centrality

    def get_lppo(self, count: int = 10, depth: int = 1):
        cell_centrality = self.get_centrality(depth=depth)
        centrality_index = sorted(range(len(cell_centrality)), key=lambda i: cell_centrality[i], reverse=True)
        return centrality_index[:count]


from setuptools import setup

setup(name='cellworld',
      description='Maciver Lab computational biology research package',
      author='german espinosa',
      author_email='germanespinosa@gmail.com',
      long_description=open('./cellworld/readme.md').read() + '\n---\n<small>Package created with Easy-pack</small>\n',
      long_description_content_type='text/markdown',
      packages=['cellworld'],
      install_requires=['numpy', 'scipy', 'matplotlib', 'json-cpp>=1.0.77', 'tcp-messages', 'networkx', 'cv'],
      license='MIT',
      include_package_data=True,
      version='0.0.395',
      zip_safe=False)



from json_cpp import JsonObject, JsonList
from .util import *


class Coordinates(JsonObject):
    """A class representing coordinates with x and y values.

    Attributes:
        x (int): The x-coordinate.
        y (int): The y-coordinate.

    Methods:
        __init__(self, x=0, y=0):
            Initializes a Coordinates object with optional x and y values.

        __hash__(self):
            Computes the hash value of the Coordinates object based on its x and y values.

        __eq__(self, other):
            Checks if two Coordinates objects are equal by comparing their x and y values.

        __add__(self, o):
            Adds two Coordinates objects and returns a new Coordinates object as the result.

        __sub__(self, o):
            Subtracts two Coordinates objects and returns a new Coordinates object as the result.

        __neg__(self):
            Negates the x and y values of the Coordinates object and returns a new Coordinates object.

        manhattan(self, o) -> int:
            Calculates the Manhattan distance between two Coordinates objects and returns an integer.

    """
    def __init__(self,
                 x: int = 0,
                 y: int = 0):
        self.x = int(x)
        self.y = int(y)

    def __hash__(self):
        hash((self.x, self.y))

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y

    def __add__(self, o):
        c = Coordinates()
        c.x = self.x + o.x
        c.y = self.y + o.y
        return c

    def __sub__(self, o):
        c = Coordinates()
        c.x = self.x - o.x
        c.y = self.y - o.y
        return c

    def __neg__(self):
        c = Coordinates()
        c.x = -self.x
        c.y = -self.y
        return c

    def manhattan(self, o) -> int:
        """Calculate the Manhattan distance between two Coordinates objects.

        Args:
            o (Coordinates): The other Coordinates object.

        Returns:
            int: The Manhattan distance as an integer.
        """
        return abs(o.x - self.x) + abs(o.y - self.y)


class Coordinates_list (JsonList):
    """A list-like class for storing and manipulating lists of Coordinates objects.

    Attributes:
        iterable (iterable): An iterable to initialize the list with.

    Methods:
        __init__(self, iterable=None):
            Initializes a Coordinates_list object with an optional iterable of Coordinates objects.

        get_x(self):
            Extracts the x values from all Coordinates objects in the list and returns them as a list.

        get_y(self):
            Extracts the y values from all Coordinates objects in the list and returns them as a list.

    """
    def __init__(self, iterable=None):
        JsonList.__init__(self, iterable=iterable, list_type=Coordinates)

    def get_x(self):
        """Get the x values of all Coordinates objects in the list.

        Returns:
            list: A list of x values.
        """
        x = []
        for coordinates in self:
            x.append(coordinates.x)
        return x

    def get_y(self):
        """Get the y values of all Coordinates objects in the list.

        Returns:
            list: A list of y values.
        """
        y = []
        for coordinates in self:
            y.append(coordinates.y)
        return y

